/*
 * Broadcom chipcommon NAND flash interface
 *
 * Copyright (C) 2010, Broadcom Corporation. All Rights Reserved.
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: nflash.c,v 1.4.4.2 2010-11-10 04:19:38 Exp $
 */

#include <typedefs.h>
#include <osl.h>
#include <bcmutils.h>
#include <siutils.h>
#include <hndsoc.h>
#include <sbhndcpu.h>
#include <sbchipc.h>
#include <bcmdevs.h>
#include <nflash.h>

#ifdef BCMDBG
#define	NFL_MSG(args)	printf args
#else
#define	NFL_MSG(args)
#endif	/* BCMDBG */

/* Private global state */
static struct nflash nflash;

/* Issue a nand flash command */
static INLINE void
nflash_cmd(osl_t *osh, chipcregs_t *cc, uint opcode)
{
	W_REG(osh, &cc->nand_cmd_start, opcode);
	/* read after write to flush the command */
	R_REG(osh, &cc->nand_cmd_start);
}

static bool firsttime = TRUE;

/* Initialize nand flash access */
struct nflash *
nflash_init(si_t *sih, chipcregs_t *cc)
{
	uint32 id, id2;
	char *name = "";
	osl_t *osh;
	int i;
	uint32 ncf, val;

	ASSERT(sih);
	/* Only support chipcommon revision >= 38 for now */
	/* chipc revision 39 does not contain the NAND flash interface */
	if ((sih->ccrev < 38) || (sih->ccrev == 39))
		return NULL;

	if (!firsttime && nflash.size)
		return &nflash;

	osh = si_osh(sih);
	bzero(&nflash, sizeof(nflash));
	/* Read flash id */
	nflash_cmd(osh, cc, NCMD_ID_RD);
	if (nflash_poll(sih, cc) < 0)
		return NULL;
	id = R_REG(osh, &cc->nand_devid);
	id2 = R_REG(osh, &cc->nand_devid_x);
	for (i = 0; i < 5; i++) {
		if (i < 4)
			nflash.id[i] = (id >> (8*i)) & 0xff;
		else
			nflash.id[i] = id2 & 0xff;
	}
	nflash.type = nflash.id[0];
	switch (nflash.type) {
	case NFL_VENDOR_AMD:
		name = "AMD";
		break;
	case NFL_VENDOR_NUMONYX:
		name = "Numonyx";
		break;
	case NFL_VENDOR_MICRON:
		name = "Micron";
		break;
	case NFL_VENDOR_TOSHIBA:
		name = "Toshiba";
		break;
	case NFL_VENDOR_HYNIX:
		name = "Hynix";
		break;
	case NFL_VENDOR_SAMSUNG:
		name = "Samsung";
		break;
	}
	ncf = R_REG(osh, &cc->nand_config);
	/*  Page size (# of bytes) */
	val = (ncf & NCF_PAGE_SIZE_MASK) >> NCF_PAGE_SIZE_SHIFT;
	switch (val) {
	case 0:
		nflash.pagesize = 512;
		break;
	case 1:
		nflash.pagesize = (1 << 10) * 2;
		break;
	case 2:
		nflash.pagesize = (1 << 10) * 4;
		break;
	case 3:
		nflash.pagesize = (1 << 10) * 8;
		break;
	}
	/* Block size (# of bytes) */
	val = (ncf & NCF_BLOCK_SIZE_MASK) >> NCF_BLOCK_SIZE_SHIFT;
	switch (val) {
	case 0:
		nflash.blocksize = (1 << 10) * 16;
		break;
	case 1:
		nflash.blocksize = (1 << 10) * 128;
		break;
	case 2:
		nflash.blocksize = (1 << 10) * 8;
		break;
	case 3:
		nflash.blocksize = (1 << 10) * 512;
		break;
	case 4:
		nflash.blocksize = (1 << 10) * 256;
		break;
	default:
		printf("Unknown block size\n");
		return NULL;
	}
	/* NAND flash size in MBytes */
	val = (ncf & NCF_DEVICE_SIZE_MASK) >> NCF_DEVICE_SIZE_SHIFT;
	if (val == 0) {
		printf("Unknown flash size\n");
		return NULL;
	}
	nflash.size = (1 << (val - 1)) * 8;
	nflash.numblocks = (nflash.size * (1 << 10)) / (nflash.blocksize >> 10);
	if (firsttime)
		printf("Found a %s NAND flash with %uB pages or %dKB blocks; total size %dMB\n",
		       name, nflash.pagesize, (nflash.blocksize >> 10), nflash.size);
	firsttime = FALSE;
	return nflash.size ? &nflash : NULL;
}

/* Read len bytes starting at offset into buf. Returns number of bytes read. */
int
nflash_read(si_t *sih, chipcregs_t *cc, uint offset, uint len, uchar *buf)
{
	uint32 mask;
	osl_t *osh;
	int i;
	uint32 *to;
	uint32 val;
	uint res;

	ASSERT(sih);
	mask = NFL_SECTOR_SIZE - 1;
	if ((offset & mask) != 0 || (len & mask) != 0)
		return 0;
	if ((((offset + len) >> 20) > nflash.size) ||
	    ((((offset + len) >> 20) == nflash.size) &&
	     (((offset + len) & ((1 << 20) - 1)) != 0)))
		return 0;
	osh = si_osh(sih);
	to = (uint32 *)buf;
	res = len;
	while (res > 0) {
		W_REG(osh, &cc->nand_cmd_addr, offset);
		nflash_cmd(osh, cc, NCMD_PAGE_RD);
		if (nflash_poll(sih, cc) < 0)
			break;
		if (((val = R_REG(osh, &cc->nand_intfc_status)) & NIST_CACHE_VALID) == 0)
			break;
		W_REG(osh, &cc->nand_cache_addr, 0);
		for (i = 0; i < NFL_SECTOR_SIZE; i += 4, to++) {
			*to = R_REG(osh, &cc->nand_cache_data);
		}
		res -= NFL_SECTOR_SIZE;
		offset += NFL_SECTOR_SIZE;
	}
	return (len - res);
}

#define NF_RETRIES	1000000

/* Poll for command completion. Returns zero when complete. */
int
nflash_poll(si_t *sih, chipcregs_t *cc)
{
	osl_t *osh;
	int i;
	uint32 pollmask;

	ASSERT(sih);
	osh = si_osh(sih);
	pollmask = NIST_CTRL_READY|NIST_FLASH_READY;
	for (i = 0; i < NF_RETRIES; i++) {
		if ((R_REG(osh, &cc->nand_intfc_status) & pollmask) == pollmask) {
			return 0;
		}
	}
	printf("nflash_poll: not ready\n");
	return -1;
}

/* Write len bytes starting at offset into buf. Returns number of bytes
 * written.
 */
int
nflash_write(si_t *sih, chipcregs_t *cc, uint offset, uint len, const uchar *buf)
{
	uint32 mask;
	osl_t *osh;
	int i;
	uint32 *from;
	uint res;
	uint32 reg;

	ASSERT(sih);
	mask = nflash.pagesize - 1;
	/* Check offset and length */
	if ((offset & mask) != 0 || (len & mask) != 0)
		return 0;
	if ((((offset + len) >> 20) > nflash.size) ||
	    ((((offset + len) >> 20) == nflash.size) &&
	     (((offset + len) & ((1 << 20) - 1)) != 0)))
		return 0;
	osh = si_osh(sih);
	/* disable partial page enable */
	reg = R_REG(osh, &cc->nand_acc_control);
	reg &= ~NAC_PARTIAL_PAGE_EN;
	W_REG(osh, &cc->nand_acc_control, reg);

	from = (uint32 *)buf;
	res = len;
	while (res > 0) {
		W_REG(osh, &cc->nand_cache_addr, 0);
		for (i = 0; i < nflash.pagesize; i += 4, from++) {
			if (i % 512 == 0)
				W_REG(osh, &cc->nand_cmd_addr, i);
			W_REG(osh, &cc->nand_cache_data, *from);
		}
		W_REG(osh, &cc->nand_cmd_addr, offset + nflash.pagesize - 512);
		nflash_cmd(osh, cc, NCMD_PAGE_PROG);
		if (nflash_poll(sih, cc) < 0)
			break;
		res -= nflash.pagesize;
		offset += nflash.pagesize;
	}
	return (len - res);
}

/* Erase a region. Returns number of bytes scheduled for erasure.
 * Caller should poll for completion.
 */
int
nflash_erase(si_t *sih, chipcregs_t *cc, uint offset)
{
	osl_t *osh;

	ASSERT(sih);

	osh = si_osh(sih);
	if ((offset >> 20) >= nflash.size)
		return -1;
	if ((offset & (nflash.blocksize - 1)) != 0) {
		return -1;
	}
	W_REG(osh, &cc->nand_cmd_addr, offset);
	nflash_cmd(osh, cc, NCMD_BLOCK_ERASE);
	if (nflash_poll(sih, cc) < 0)
		return -1;
	return 0;
}

int nflash_checkbadb(si_t *sih, chipcregs_t *cc, uint offset)
{
	osl_t *osh;
	int i;
	uint off;

	ASSERT(sih);

	osh = si_osh(sih);
	if ((offset >> 20) >= nflash.size)
		return -1;
	if ((offset & (nflash.blocksize - 1)) != 0) {
		return -1;
	}
	for (i = 0; i < 2; i++) {
		off = offset + nflash.pagesize;
		W_REG(osh, &cc->nand_cmd_addr, off);
		nflash_cmd(osh, cc, NCMD_SPARE_RD);
		if (nflash_poll(sih, cc) < 0)
			return -1;
		if ((R_REG(osh, &cc->nand_intfc_status) & NIST_SPARE_VALID) != NIST_SPARE_VALID)
			return -1;
		if ((R_REG(osh, &cc->nand_spare_rd0) & 0xff) != 0xff)
			return -1;
	}
	return 0;
}
